/*
 *                   jfjji,.           H A M M E R H E A D  S H A R K W A R E
 *                  ,#######EL:                  http://www.hhsw.de
 *                   .,j######Kj       
 *                    ;W########G:     This  code  is  free software; you can
 *                   j#####G ,D#L:     redistribute it and/or modify it under
 *             .,;;,G#####D.           the  terms  of  the GNU General Public
 *           jE##########G             License   as  published  by  the  Free
 *             .,W######D.             Software  Foundation; either version 2
 *             :L#####K##G             of  the  License,  or (at your option)
 *           .f#####G;.t##:            any later version.
 *         ;L###ELi.    ,D,            
 *       iEEGWj                        This  code  is distributed in the hope
 *      tW;                            that  it  will  be useful, but without
 *     jf                              any warranty; without even the implied
 *    .Kt                              warranty of merchantability or fitness
 *    i:                               for  a particular purpose. See the GNU
 *    ;                                General   Public   License   for  more
 *    .                                details (http://www.gnu.org).
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <iostream>
#include <string>

#include "Raise.hh"
#include "Slot.hh"

extern char **environ;

using namespace std;
using namespace bbdock;

/**
 * Initialize object
 *
 * @param d - Dock object
 * @param i - Icon object
 * @param n - next Slot in chain
 */
Slot::Slot( Dock &d, Icon &i, Slot *n ) : 
	dock( &d ), 
	icon( &i ), 
	next( n ),
	render( 0 ),
	iconimage( 0 ),
	surface( 0 ),
	background( 0 ),
	invalid( false ),
	visibility( VisibilityUnobscured ),
	pid( 0 ),
	transparency( Full )
{
	loadIcon();

	Window	root;
	Window	dummy;

	root = RootWindow( dock->getDisplay(), dock->getScreen() );

	// create dummy window and dock slot
	{
		dummy = XCreateSimpleWindow( dock->getDisplay(), root, 
			0, 0, 1, 1, 1, 0, 0 );

		XSizeHints	sh;

	    sh.x = 0;
		sh.y = 0;
		sh.width = dock->getSlotWidth();
		sh.height = dock->getSlotHeight();

	    XSetWindowAttributes	xswat;
		unsigned long			vmask;

		xswat.background_pixmap = ParentRelative;
		vmask = CWBackPixmap;

		window = XCreateWindow( dock->getDisplay(), root, 
			sh.x, sh.y, 
			sh.width, sh.height,
			0, 
			CopyFromParent, 
			InputOutput,
			CopyFromParent, 
			vmask, &xswat );

		sh.flags = USSize | USPosition;
		XSetWMNormalHints( dock->getDisplay(), dummy, &sh );

		{
			XWMHints	xwmh;

			xwmh.initial_state = WithdrawnState;
			xwmh.icon_window = window;
			xwmh.icon_x = sh.x;
			xwmh.icon_y = sh.y;
			xwmh.window_group = dummy;
			xwmh.flags = StateHint | IconWindowHint |
				IconPositionHint | WindowGroupHint;

			XSetWMHints( dock->getDisplay(), dummy, &xwmh );
		}
	}

	// set name
	{
		XClassHint	xch;

		xch.res_name = "bbdock";
		xch.res_class = "bbdock";

		XSetClassHint( dock->getDisplay(), dummy, &xch );
	}

	XSelectInput( dock->getDisplay(), window,
		ButtonReleaseMask | ButtonPressMask |
		ExposureMask | VisibilityChangeMask | 
		StructureNotifyMask | SubstructureNotifyMask | 
		PropertyChangeMask );

	XMapWindow( dock->getDisplay(), dummy );
}

/**
 * Clean up
 */
Slot::~Slot()
{
	destorySurface();
	destroyIcon();
}

/**
 * Display icon
 *
 * @param refresh - refresh background
 */
void Slot::draw( bool refresh )
{
	if( refresh || invalid )
	{
		if( invalid )
			getBackground();
		else
			memcpy( image, background, render->getSize() );

		if( transparency < Full )
			render->ghosted( (unsigned char *) image, 
				(unsigned char *) iconimage, transparency );
		else if( pid && !icon->getTitle().empty() )
			render->active( (unsigned char *) image, 
				(unsigned char *) iconimage );
		else
			render->opaque( (unsigned char *) image, 
				(unsigned char *) iconimage );
	}

	XPutImage( dock->getDisplay(), window, dock->getGC(), surface, 0, 0, 
		left, top, width, height );
}

/**
 * Execute command
 */
void Slot::exec()
{
	// raise window if already running
	if( !icon->getTitle().empty() && pid )
	{
		Raise	r( dock->getDisplay() );

		r.raise( icon->getTitle().c_str() );
		return;
	}

	if( transparency < Full ||
		icon->getCommand().empty() )
		return;

	switch( (pid = fork()) )
	{
		case -1:
			throw "Fork failed !";
			return;
		case 0:
			{
				char *argv[4];

				argv[0] = "sh";
				argv[1] = "-c";
				argv[2] = (char *) icon->getCommand().c_str();
				argv[3] = 0;

				setsid();
				execve( "/bin/sh", argv, environ );

				throw "Exec failed !";
			}
			return;
	}

	transparency = Half;
	draw( true );
}

/**
 * Process timer event
 */
void Slot::tick()
{
	int	status;

//	if( !icon->getTitle().empty() && pid &&
	if( pid &&
		waitpid( pid, &status, WNOHANG | WUNTRACED ) == pid )
	{
		pid = 0;
		transparency = Full;

		draw( true );
	}

	if( //icon->getTitle().empty() && 
		transparency < Full )
	{
		transparency = (int) ((double) transparency*1.05);
		
		if( transparency > Full )
			transparency = Full;

		draw( true );
	}
}

/**
 * (Re-)Set icon image
 *
 * @param filename - image file to load
 */
void Slot::setIcon( const char *filename )
{
	icon->setImage( filename );
	loadIcon();
	invalidate();
	draw( true );
}

/**
 * Load PNG icon
 */
void Slot::loadIcon()
{
	destroyIcon();
	width = height = 0;

	png_structp	png = 0;
	png_infop	info = 0;
	FILE		*fp = 0;

	if( !(png = png_create_read_struct( PNG_LIBPNG_VER_STRING,
		0, 0, 0 )) )
		throw "PNG library error !";

	try
	{
		if( !(info = png_create_info_struct( png )) ||
			setjmp( png_jmpbuf( png ) ) ||
			!(fp = fopen( icon->getImage().c_str(), "rb" )) )
			throw 0;

		png_init_io( png, fp );
		png_set_sig_bytes( png, 0 );
		png_read_png( png, info, 
			PNG_TRANSFORM_IDENTITY | 
			PNG_TRANSFORM_EXPAND |
			PNG_TRANSFORM_BGR,
			0 );

		if( !(info->valid & PNG_INFO_IDAT) ||
			!(iconimage = new int[info->width*info->height]) )
			throw 0;

		for( int y = 0, *src, *dest = iconimage; 
			y < info->height && (src = (int *) info->row_pointers[y]); 
			y++, dest += info->width )
			memcpy( dest, src, info->width<<2 );

		width = info->width;
		height = info->height;

		// convert grayscale image to rgb
		if( info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA )
			for( int y = info->height, *line = iconimage;
				--y; line += info->width )
			{
				unsigned char *dest = (unsigned char *) line+(info->width<<2);
				unsigned char *src = (unsigned char *) line+(info->width<<1);

				for( int x = info->width; --x; )
				{
					*(--dest) = *(--src);
					*(--dest) = *(--src);
					*(--dest) = *src;
					*(--dest) = *src;
				}
			}
	}
	catch( int )
	{
	}

	// free possibly allocated resources
	if( fp )
		fclose( fp );

	png_destroy_read_struct( &png, &info, (png_infopp) 0 );

	if( !width || !height )
		throw "Invalid image file !";

	if( (left = (dock->getSlotWidth()-width)>>1) < 0 ||
		(top = (dock->getSlotHeight()-height)>>1) < 0 )
		throw "Image is too big !";

	// (re-)create surface for this image size
	createSurface();
}

/**
 * Free image resource
 */
void Slot::destroyIcon()
{
	if( !iconimage )
		return;

	delete iconimage;
	iconimage = 0;
}

/**
 * (Re-)Create the surface
 */
void Slot::createSurface()
{
	destorySurface();

	if( !(render = Render::getInstance( dock->getBitsPerPixel(), 
			width, height )) ||
		!(background = (char *) calloc( render->getSize(), sizeof( char ) )) ||
		!(image = (char *) calloc( render->getSize(), sizeof( char ) )) ||
		!(surface = XCreateImage( dock->getDisplay(), 
			DefaultVisual( dock->getDisplay(), dock->getScreen() ), 
			DefaultDepth( dock->getDisplay(), dock->getScreen() ), 
			ZPixmap, 0, image, 
			width, height, 
			32, 0 )) )
		throw "Could not create XImage !";
}

/**
 * Drop surface
 */
void Slot::destorySurface()
{
	if( !surface )
		return;

	XDestroyImage( surface );
	surface = 0;
	// bitmap memory gets free'd automatically

	if( background )
	{
		free( background );
		background = 0;
	}
	
	if( render )
	{
		delete render;
		render = 0;
	}
}

/**
 * Get background image from parent window
 */
void Slot::getBackground()
{
	// ensure background is marked as invalid
	invalid = true;

	Window			slit;

	// get parent window
	{
		Window			root;
		Window			*childs;
		unsigned int	n;

 		XQueryTree( dock->getDisplay(), window, &root, &slit, &childs, &n );

		if( n && childs )
			XFree( childs );
	}

	if( !slit )
		return;			

	// check if window is visible on screen since XGetSubImage 
	// fails otherwise
	{
		XWindowAttributes	wa;

		XGetWindowAttributes( dock->getDisplay(), slit, &wa );

		if( wa.map_state != IsViewable ||
			wa.x < 0 ||
			wa.y < 0 || 
			wa.x+wa.width > dock->getScreenWidth() || 
			wa.y+wa.height > dock->getScreenHeight() )
			return;
	}

	XClearWindow( dock->getDisplay(), window );
	XGetSubImage( dock->getDisplay(), window, left, top, width, height, 
		0xffffffff, ZPixmap, surface, 0, 0 );

	if( visibility == VisibilityUnobscured )
	{
		memcpy( background, image, render->getSize() );

		invalid = false;
	}
}
